/*!
 * module-loader.js v0.0.7
 * (c) 2014-2020 stwilson
 * Released under the MIT License.
 */
import { Container, injectable, inject, METADATA_KEY, decorate, multiInject } from 'inversify';
export { inject, injectable } from 'inversify';
import 'reflect-metadata';

class IocManager {
    constructor() {
        this.denpendcyRegisters = [];
        this.diConainter = new Container();
    }
    addDenpendcyRegistrar(registrar) {
        this.denpendcyRegisters.push(registrar);
    }
    registerModuleByConvention(moduleContexts) {
        let registerCTX = new RegistrationContext(moduleContexts, this);
        for (let register of this.denpendcyRegisters) {
            register.register(registerCTX);
        }
    }
    register(ContainerDelegate) {
        if (ContainerDelegate !== undefined)
            ContainerDelegate(this.diConainter);
    }
    get(serviceIdentifier) {
        return this.diConainter.get(serviceIdentifier);
    }
    getAll(serviceIdentifier) {
        return this.diConainter.getAll(serviceIdentifier);
    }
    isRegistered(serviceIdentifier) {
        return this.diConainter.isBound(serviceIdentifier);
    }
    unbindAll() {
        this.diConainter.unbindAll();
    }
    getContainer() {
        return this.diConainter;
    }
}
class RegistrationContext {
    constructor(moduleContexts, iocManager) {
        this.iocManger = iocManager;
        this.moduleContexts = moduleContexts;
    }
}

class BootstrapperOptions {
    constructor() {
        this.iocManager = new IocManager();
    }
}

const Types = {
    IBlocksShell: Symbol.for("IBlocksShell"),
    IRouteProvider: Symbol.for("IRouteProvider"),
    ITemplateProvider: Symbol.for("ITemplateProvider"),
    IBootstrapper: Symbol.for("IBootstrapper"),
    IManifestProvider: Symbol.for("IManifestProvider"),
    IResourceManager: Symbol.for("IResourceManager"),
    IRouteManager: Symbol.for("IRouteManager"),
};

/*! *****************************************************************************
Copyright (c) Microsoft Corporation. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License"); you may not use
this file except in compliance with the License. You may obtain a copy of the
License at http://www.apache.org/licenses/LICENSE-2.0

THIS CODE IS PROVIDED ON AN *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
MERCHANTABLITY OR NON-INFRINGEMENT.

See the Apache Version 2.0 License for specific language governing permissions
and limitations under the License.
***************************************************************************** */

function __decorate(decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
}

function __param(paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
}

function __metadata(metadataKey, metadataValue) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(metadataKey, metadataValue);
}

// let _win: any = window;
// let globalIocManager: IocManager = _win["globalIocManager"] || new IocManager();
// _win["globalIocManager"] = globalIocManager;
let IDependency = class IDependency {
};
IDependency = __decorate([
    injectable()
], IDependency);
class IBlocksShell {
    constructor() {
        this.types = [];
        this.typeMapModuleName = new Map();
        this.moduleMapTypes = new Map();
    }
    async initialize() {
        throw new Error("initialize is not implemented. ");
    }
    get BlocksModules() {
        throw new Error("initialize is not implemented. ");
    }
}
class IBootstrapper {
    get PlugInSources() {
        throw new Error("plugInSources is not implemented. ");
    }
    initialize() {
        throw new Error("initialize is not implemented. ");
    }
    dispose() {
        throw new Error("dispose is not implemented. ");
    }
}
class IShell extends IDependency {
    constructor() {
        super(...arguments);
        this.pluginSource = [];
        this.types = [];
        this.moduleMapTypes = new Map();
        this.typeMapModuleName = new Map();
    }
    initialize() {
    }
}
class BlocksModule extends IDependency {
    constructor() {
        super();
        //TODO Need to SortModule??
        this.order = 10;
        this.providers = () => [];
        this.moduleName = "default";
    }
    async preInitialize() {
        // let tsContext = require.context("./", true);
        // this.iocManager.RegisterModuleByConvention(this.providers);
    }
    async initialize() {
    }
    getProviders() {
        return this.providers();
    }
    async dispose() {
    }
}
__decorate([
    inject(IocManager),
    __metadata("design:type", IocManager)
], BlocksModule.prototype, "iocManager", void 0);

class IRouteProvider extends IDependency {
    getRoutes() {
        throw new Error("getRoutes is not implemented.");
    }
}
class ITemplateProvider extends IDependency {
    getTemplate() {
        throw new Error("getTemplate is not implemented.");
    }
}
class RouteResult {
    constructor() {
        this.layout = "";
        this.path = "";
        this.children = [];
    }
}
class TemplateResult {
}

class IManifestProvider extends IDependency {
    buildManifests(builder) {
        throw new Error("buildManifests is not implemented.");
    }
}
class ResourceManifest {
    constructor() {
        this.resources = [];
    }
    defineResource(resourceType, resourceName) {
        let manifest = new ResourceDefinition(resourceType, resourceName);
        this.resources.push(manifest);
        return manifest;
    }
}
class ResourceDefinition {
    constructor(resourceType, resourceName) {
        this.resourceName = resourceName;
        this.resourceType = resourceType;
    }
    add(resource) {
        this.resource = resource;
    }
}
class ResourceManifestBuilder {
    constructor() {
        this.resourceManifests = [];
    }
    add() {
        let manifest = new ResourceManifest();
        this.resourceManifests.push(manifest);
        return manifest;
    }
}

let decorateIfNoExist = (decorator, target, parameterIndex) => {
    if (Reflect.hasOwnMetadata(METADATA_KEY.PARAM_TYPES, target))
        return;
    return decorate(decorator, target, parameterIndex);
};

let RouteManager = class RouteManager {
    constructor(IocManager) {
        this.iocManager = IocManager;
    }
    getUniquePath() {
        let routes = this.getRoute();
        let routePaths = [];
        for (const route of routes) {
            if (!route.children) {
                routePaths.push(Object.assign(true, {}, route, { meta: { uniqueKey: route.uniqueKey } }));
                continue;
            }
            for (let routeChild of route.children) {
                if (!routeChild.children) {
                    routePaths.push(Object.assign(true, routeChild, { path: route.path + "/" + routeChild.path, meta: { uniqueKey: routeChild.uniqueKey } }));
                    continue;
                }
                for (const routeChildTemp of routeChild.children) {
                    if (!routeChildTemp.children) {
                        routePaths.push(Object.assign(true, routeChild, { path: route.path + "/" + routeChild.path + "/" + routeChildTemp.path, meta: { uniqueKey: routeChildTemp.uniqueKey } }));
                        continue;
                    }
                }
            }
        }
        return routePaths;
    }
    getRoute() {
        let routeObj = [];
        let templateProviders = this.iocManager.isRegistered(Types.ITemplateProvider) ? this.iocManager.getAll(Types.ITemplateProvider) : [];
        if (!this.iocManager.isRegistered(Types.IRouteProvider))
            return routeObj;
        for (let routeProvider of this.iocManager.getAll(Types.IRouteProvider)) {
            let moduleName = this.iocManager.get(Types.IBlocksShell)
                .typeMapModuleName.get(routeProvider.constructor);
            let routes = routeProvider.getRoutes();
            if (!routes || !Array.isArray(routes))
                continue;
            let layoutRouteChilds = {};
            for (const route of routes) {
                route.path = moduleName + "/" + route.path;
                if (route.uniqueKey) {
                    route.uniqueKey = RouteManager.getUniqueKey(moduleName, route.uniqueKey); //  route.uniqueKey + "_" + moduleName;
                    route.meta = Object.assign(true, route.meta, { uniqueKey: route.uniqueKey });
                }
                if (route.layout) {
                    layoutRouteChilds[route.layout] = layoutRouteChilds[route.layout] ? layoutRouteChilds[route.layout] : new RouteResult();
                    let tempRoute = layoutRouteChilds[route.layout];
                    for (const templateProvider of templateProviders) {
                        let template = templateProvider.getTemplate().find((t) => t.name === route.layout);
                        if (template) {
                            tempRoute = Object.assign(tempRoute, template);
                            route.name = route.name ? route.name : route.uniqueKey;
                            tempRoute.children.push(route);
                            break;
                        }
                    }
                    if (tempRoute.children.length < 1)
                        routeObj.push(route);
                    continue;
                }
                if (!route.path.startsWith("/"))
                    route.path = "/" + route.path;
                routeObj.push(route);
            }
            for (let layoutRouteKey in layoutRouteChilds) {
                if (layoutRouteChilds.hasOwnProperty(layoutRouteKey)) {
                    const tempRoute = layoutRouteChilds[layoutRouteKey];
                    routeObj.push(Object.assign({}, tempRoute));
                }
            }
        }
        return routeObj;
    }
};
RouteManager.getUniqueKey = (moduleName, uniqueKey) => uniqueKey + "_" + moduleName;
RouteManager = __decorate([
    __param(0, inject(IocManager)),
    __metadata("design:paramtypes", [IocManager])
], RouteManager);
var DefaultRouteManager = RouteManager;

class RouteStartupModule extends BlocksModule {
    constructor() {
        super();
        this.moduleName = "RouteStartupModule";
    }
    async preInitialize() {
        if (!this.shell)
            throw new Error("Shell is null or empty.");
        if (!this.iocManager)
            throw new Error("iocManager is null or empty.");
        let iocManagerTmp = this.iocManager;
        iocManagerTmp.register((c) => {
            //hahhahaahahah
            debugger;
            c.bind(Types.IRouteManager).to(DefaultRouteManager).inSingletonScope();
            decorateIfNoExist(injectable(), DefaultRouteManager);
        });
        this.shell.moduleMapTypes.forEach((typeMap, ModuleType) => {
            typeMap.forEach((type, index) => {
                if (type.prototype instanceof IRouteProvider) {
                    iocManagerTmp.register((c) => {
                        c.bind(Types.IRouteProvider).to(type).inTransientScope();
                    });
                }
                if (type.prototype instanceof ITemplateProvider) {
                    iocManagerTmp.register((c) => {
                        c.bind(Types.ITemplateProvider).to(type).inTransientScope();
                    });
                }
            });
        });
    }
}
__decorate([
    inject(Types.IBlocksShell),
    __metadata("design:type", IShell)
], RouteStartupModule.prototype, "shell", void 0);

let typeCheckExpression = [];

let resourceManager = class resourceManager {
    constructor(blocksShell, manifestProviders) {
        this.resources = new Map();
        this.blocksShell = blocksShell;
        this.manifestProviders = manifestProviders;
    }
    getResources() {
        if (this.resources.size === 0)
            this.buildResource();
        return this.resources;
    }
    getResource(moduleName, resourceType, resourceName) {
        let moduleResource = this.getResources().get(moduleName);
        if (!moduleResource || !moduleResource.has(resourceManager.getResourceKey(resourceType, resourceName))) {
            throw new Error(`Can not found resourceType:${resourceType}, resourceName:${resourceName} in resources.`);
        }
        return moduleResource.get(resourceManager.getResourceKey(resourceType, resourceName));
    }
    buildResource() {
        for (const manifestProvider of this.manifestProviders) {
            let builder = new ResourceManifestBuilder();
            manifestProvider.buildManifests(builder);
            let moduleName = this.blocksShell.typeMapModuleName.get(manifestProvider.constructor);
            if (moduleName === undefined)
                throw new Error(`${manifestProvider.constructor} can not found moduleName.`);
            if (!this.resources.has(moduleName))
                this.resources.set(moduleName, new Map());
            for (const manifest of builder.resourceManifests) {
                for (const resource of manifest.resources) {
                    let resourceNameMap = this.resources.get(moduleName);
                    if (resourceNameMap)
                        resourceNameMap.set(resourceManager.getResourceKey(resource.resourceType, resource.resourceName), resource);
                }
            }
        }
    }
};
resourceManager.getResourceKey = (resourceType, resourceName) => resourceType + "-" + resourceName;
resourceManager = __decorate([
    __param(0, inject(Types.IBlocksShell)), __param(1, multiInject(Types.IManifestProvider)),
    __metadata("design:paramtypes", [IBlocksShell, Array])
], resourceManager);

class manifestStartupModule extends BlocksModule {
    constructor() {
        super();
        this.moduleName = "ManifestStartupModule";
    }
    async preInitialize() {
        if (!this.shell)
            throw new Error("Shell is null or empty.");
        if (!this.iocManager)
            throw new Error("iocManager is null or empty.");
        let iocManagerTmp = this.iocManager;
        iocManagerTmp.register((c) => {
            c.bind(Types.IResourceManager).to(resourceManager).inSingletonScope();
            decorateIfNoExist(injectable(), resourceManager);
        });
        this.shell.moduleMapTypes.forEach((typeMap, ModuleType) => {
            typeMap.forEach((type, index) => {
                if (type.prototype instanceof IManifestProvider) {
                    iocManagerTmp.register((c) => {
                        c.bind(Types.IManifestProvider).to(type).inTransientScope();
                    });
                }
            });
        });
    }
}
__decorate([
    inject(Types.IBlocksShell),
    __metadata("design:type", IShell)
], manifestStartupModule.prototype, "shell", void 0);

let BlocksShell = class BlocksShell {
    constructor(iocManager) {
        this.pluginSource = [];
        this.iocManager = iocManager;
        this.types = [];
        this.typeMapModuleName = new Map();
        this.moduleMapTypes = new Map();
        this.typeMapFileName = new Map();
        this.blocksModules = [];
    }
    get BlocksModules() {
        return this.blocksModules;
    }
    async initialize() {
        for (let filesArray of this.getTypeFromFiles()) {
            this.moduleAndDependencyLoader(filesArray);
        }
        this.autoRegisterModule(RouteStartupModule);
        this.autoRegisterModule(manifestStartupModule);
        let startupModules = this.iocManager.getAll(BlocksModule);
        this.blocksModules = startupModules;
        //
        let temp = new Map();
        for (let module of startupModules) {
            if (temp.has(module.moduleName))
                throw new Error(`ModuleName ${module.moduleName} is registered double.`);
            temp.set(module.moduleName, []);
        }
        // for (let module of startupModules) {
        //     if (this.ModuleMapTypes.has(module.ModuleName))
        //         throw new Error(`ModuleName {ModuleName} is registered double.`)
        //     this.ModuleMapTypes.set(module.ModuleName, []);
        // }
        this.moduleMapTypes.forEach((moduleMapSet, moduleTypeKey) => {
            let moduleType = this.typeMapFileName.get(moduleTypeKey);
            if (!moduleType)
                return;
            let moduleFilePrefix = moduleType[1];
            moduleFilePrefix = moduleFilePrefix.substring(0, moduleFilePrefix.lastIndexOf('/') + 1);
            this.typeMapFileName.forEach((value, typeMapKey) => {
                if (value[1].startsWith(moduleFilePrefix)) {
                    moduleMapSet.push(typeMapKey);
                }
            });
        });
        for (const startupModule of startupModules) {
            if (!this.moduleMapTypes.has(startupModule.constructor))
                continue;
            let moduleType = this.moduleMapTypes.get(startupModule.constructor);
            if (!moduleType)
                continue;
            for (const type of moduleType) {
                this.typeMapModuleName.set(type, startupModule.moduleName);
            }
        }
        let orders = startupModules.map(m => m.order).reduce((cur, next) => {
            if (!cur.includes(next))
                cur.push(next);
            return cur;
        }, new Array()).sort((a, b) => a - b);
        let orderGroup = orders.reduce((cur, next) => {
            cur.push(startupModules.filter(m => m.order === next));
            return cur;
        }, new Array());
        for (const orderModules of orderGroup) {
            await Promise.all(orderModules.map(m => m.preInitialize()));
        }
        for (const orderModules of orderGroup) {
            await Promise.all(orderModules.map(m => m.initialize()));
        }
    }
    moduleAndDependencyLoader(filesArray) {
        let registerAction = (fileKey) => {
            let exportTypes = filesArray(fileKey);
            for (let objKey of Object.keys(exportTypes)) {
                let exportType = exportTypes[objKey];
                if (!exportType)
                    continue;
                let isModule = this.autoRegisterModule(exportType);
                if (isModule)
                    this.moduleMapTypes.set(exportType, []);
                if (exportType.prototype instanceof IDependency) {
                    this.typeMapFileName.set(exportType, [isModule, fileKey]);
                    this.types.push(exportTypes[objKey]);
                }
                for (const checkExppression of typeCheckExpression) {
                    let type = checkExppression(exportType);
                    if (type) {
                        this.typeMapFileName.set(type, [isModule, fileKey]);
                        this.types.push(type);
                        break;
                    }
                }
                // if (exportType.prototype instanceof Controller && exportType.options.type) {
                //     this.typeMapFileName.set(exportType.options.type, [isModule, fileKey]);
                //     this.types.push(exportType.options.type);
                // }
                // if(fileKey.lastIndexOf(".bl.ts"))
                // {
                //     debugger
                //     decorate(Component(), exportType);
                // }
            }
        };
        for (let fileKey of filesArray.keys()) {
            registerAction(fileKey);
        }
    }
    autoRegisterModule(type) {
        if (type.prototype instanceof BlocksModule) {
            this.iocManager.register((c) => {
                c.bind(BlocksModule).to(type).inTransientScope();
                decorateIfNoExist(injectable(), type);
            });
            return true;
        }
        return false;
    }
    getTypeFromFiles() {
        return this.pluginSource.values();
    }
};
BlocksShell = __decorate([
    __param(0, inject(IocManager)),
    __metadata("design:paramtypes", [IocManager])
], BlocksShell);

class BlocksBoostrapper extends IBootstrapper {
    constructor(startModule, optionsFunc) {
        super();
        let bootstarpperOptions = new BootstrapperOptions();
        if (optionsFunc !== undefined) {
            optionsFunc(bootstarpperOptions);
        }
        this.iocManager = bootstarpperOptions.iocManager;
        this.startModule = startModule;
        this.plugInSources = [];
        // this.RouteHelper = new RouteHelperCls(this.iocManager);
    }
    get RouteHelper() {
        return this.iocManager.get(Types.IRouteManager);
    }
    get PlugInSources() {
        return this.plugInSources;
    }
    static create(startModule, optionsFunc) {
        return new BlocksBoostrapper(startModule, optionsFunc);
    }
    async initialize() {
        //Logger
        this.registerShell();
        let shell = this.iocManager.get(Types.IBlocksShell);
        shell.pluginSource = this.plugInSources;
        await shell.initialize();
    }
    registerShell() {
        this.iocManager.register((c) => {
            c.bind(IocManager).toConstantValue(this.iocManager);
            c.bind(Types.IBlocksShell).to(BlocksShell).inSingletonScope();
            //c.bind<BlocksShell>(Types.IBlocksShell).to(BlocksShell).inSingletonScope();
            decorateIfNoExist(injectable(), IocManager);
            decorateIfNoExist(injectable(), BlocksShell);
        });
    }
    dispose() {
        this.iocManager.unbindAll();
    }
}

// const TYPES = {
//     BlocksModule: Symbol.for("BlocksModule"),
// };
decorate(injectable(), BlocksModule);
let Bootstrapper = BlocksBoostrapper.create(undefined, (o) => o.iocManager = new IocManager());

export { BlocksBoostrapper, BlocksModule, Bootstrapper, IBlocksShell, IBootstrapper, IDependency, IManifestProvider, IRouteProvider, IShell, ITemplateProvider, IocManager, RouteResult, TemplateResult, Types, decorateIfNoExist };
